from typing import Generator, Any, NoReturn, Union, Callable

from clang.cindex import Cursor, Token, TokenKind

from clangast.data_structures import StackingPipeline, Tree, Node
from clangast.algorithms.elements_parsing.utility import (
    gathering_tokens_until_border, gathering_statement, gathering_statement_open_spelling,
    filter_none, mismatch_spelling
)


def parse_error(t: Token) -> RuntimeError:
    return RuntimeError(
        f"Unexpected token; {t.cursor.translation_unit}:{t.cursor.location}:{t.cursor.kind}:{t.spelling}")


def expected_token(t: Token, spelling: str) -> NoReturn:
    if mismatch_spelling(t, spelling):
        raise parse_error(t)


class ObjectParser(object):
    def __init__(self, cursor: Cursor):
        self.cursor: Cursor = cursor
        self.tokens_pipeline: StackingPipeline = StackingPipeline(self.cursor.get_tokens())
        self.result_tree: Tree = Tree()

    def _get_statement_as_pipeline(self, source: StackingPipeline,
                                   border_spelling: str, gathering_method: Callable) -> StackingPipeline:
        return StackingPipeline(
            filter_none(gathering_method(source,
                                         border_spelling))
        )

    def _get_statement_as_string(self, source: StackingPipeline,
                                 border_spelling: str, gathering_method: Callable) -> str:
        return " ".join((
            token.spelling for token in self._get_statement_as_pipeline(source, border_spelling, gathering_method)
        ))

    def _parse_statement(self, source: StackingPipeline) -> Generator[Node, Any, None]:
        t = source.pop()
        if mismatch_spelling(t, '{'):
            source += t
            tokens = self._get_statement_as_pipeline(source, ';', gathering_statement)
        else:
            tokens = self._get_statement_as_pipeline(source, '{', gathering_statement_open_spelling)
        for node in self._parse_from(filter_none(tokens)):
            yield node

    def _parse_if_case(self, token: Token, source: StackingPipeline) -> Node:
        node = Node(token.spelling, token.cursor.kind)

        t: Token = source.pop()
        expected_token(t, '(')
        condition = self._get_statement_as_string(source, '(', gathering_statement)
        node.store(condition)

        for _node in self._parse_statement(source):
            node.add(_node)

        t = source.pop()
        if mismatch_spelling(t, "else"):
            source += t
            return node

        else_node = Node(t.spelling, t.cursor.kind)
        for _node in self._parse_statement(source):
            else_node.add(_node)
        node.add(else_node)

        return node

    def _parse_switch_case(self, token: Token, source: StackingPipeline) -> Node:
        node = Node(token.spelling, token.cursor.kind)

        t = source.pop()
        expected_token(t, '(')
        condition = self._get_statement_as_string(source, '(', gathering_statement)
        node.store(condition)

        t = source.pop()
        expected_token(t, '{')
        source += t

        for _node in self._parse_statement(source):
            node.add(_node)

        return node

    def _parse_case_in_switch(self, token: Token, source: StackingPipeline) -> Node:
        name = self._get_statement_as_string(source, ':', gathering_tokens_until_border)
        full_name = f"{token.spelling}_{name}"
        node = Node(full_name, token.cursor.kind)

        for token in source:
            spelling = token.spelling
            if spelling == "case":
                source += token
                node.store("fall_trough")
                return node
            elif spelling == "default":
                source += token
                node.store("fall_trough")
                return node
            elif spelling == "break":
                node.store("break")
                return node
            elif spelling == "return":
                node.store("return")
                return_node = self._parse_return_case(token, source)
                node.add(return_node)
                return node
            else:
                for _node in self._parse_statement(source):
                    node.add(node)

        expected_token(token, "end_of_case")

    def _parse_default_in_switch(self, token: Token, source: StackingPipeline) -> Node:
        node = Node(token.spelling, token.cursor.kind)
        t = source.pop()
        expected_token(t, ':')

        spelling = token.spelling
        if spelling == "break":
            node.store("break")
            return node
        elif spelling == "return":
            node.store("return")
            return_node = self._parse_return_case(token, source)
            node.add(return_node)
            return node
        else:
            for _node in self._parse_statement(source):
                node.add(node)

        expected_token(token, "end_of_default_case")

    def _parse_return_case(self, token: Token, source: StackingPipeline) -> Node:
        node = Node(token.spelling, token.cursor.kind)
        what = self._get_statement_as_string(source, ';', gathering_tokens_until_border)
        node.store(what)
        return node

    def _parsing_keyword(self, token: Token, source: StackingPipeline) -> Union[Node, None]:
        spelling = token.cursor.spelling
        if spelling == "if":
            return self._parse_if_case(token, source)
        elif spelling == "switch":
            return self._parse_switch_case(token, source)
        elif spelling == "case":
            return self._parse_case_in_switch(token, source)
        elif spelling == "default":
            return self._parse_default_in_switch(token, source)
        elif spelling == "return":
            return self._parse_return_case(token, source)
        else:
            return None

    def _parsing_identifier(self, token: Token, source: StackingPipeline) -> Union[Node, None]:
        node = Node(token.spelling, token.cursor.kind)
        rest_of_line = self._get_statement_as_string(source, ';', gathering_tokens_until_border)
        node.store(rest_of_line)

        return node

    def _parse_from(self, source: StackingPipeline) -> Generator[Node, Any, None]:
        for token in filter_none(source):
            kind = token.kind
            if kind == TokenKind.KEYWORD:
                yield self._parsing_keyword(token, source)
            elif kind == TokenKind.IDENTIFIER:
                yield self._parsing_identifier(token, source)

    def _parse(self) -> Generator[Node, Any, None]:
        for node in filter_none(self._parse_from(self.tokens_pipeline)):
            yield node

    def parse(self) -> NoReturn:
        root_node = Node(self.cursor.spelling, self.cursor.kind)
        object_prototype = self._get_statement_as_string(self.tokens_pipeline, '{', gathering_tokens_until_border)
        root_node.store(object_prototype)
        self.result_tree.set_root(root_node)

        for node in filter_none(self._parse()):
            self.result_tree.add(node, [])

    def name(self) -> str:
        return self.result_tree.root_name()

    def show(self, target_file, **kwargs):
        self.result_tree.show(target_file, **kwargs)



from clangast.data_structures import Tree, StackingPipeline
from clangast.algorithms import parse_file

from clangast.algorithms.tools import init_clang

default_path_to_library = "/usr/lib/libclang.so"

init_clang(default_path_to_library)

__all__ = [
    "Tree", "StackingPipeline", "parse_file"
]



from typing import Any, NoReturn, Deque, Sized, Iterator, Union, Iterable

from collections import deque


class EndlessPipeline(object):
    def __init__(self, source: Iterable[Any]):
        self.pipeline = source

    def __next__(self) -> Any:
        for element in self.pipeline:
            return element

    def __iter__(self) -> Any:
        t = next(self)
        while t is not None:
            yield t
            t = next(self)

    def pop(self) -> Any:
        return next(self)


class DefaultStack(object):
    def __init__(self):
        self.collection: Deque[Any] = deque()

    def __len__(self) -> int:
        return len(self.collection)

    def __next__(self) -> Any:
        while len(self) != 0:
            return self.collection.pop()

    def __add__(self, other: Any) -> NoReturn:
        self.collection.append(other)


class StackingPipeline(EndlessPipeline):
    def __init__(self, source: Iterable[Any], stacking_container: Iterator[Any] = None):
        super(StackingPipeline, self).__init__(source)
        if stacking_container is None:
            stacking_container = DefaultStack()
        self.stack = stacking_container

    def __next__(self) -> Any:
        while len(self.stack) != 0:
            return next(self.stack)
        return super(StackingPipeline, self).__next__()

    def __iter__(self) -> Any:
        t = next(self)
        while t is not None:
            yield t
            t = next(self)

    def __add__(self, other: Any) -> NoReturn:
        self.stack += other

    def pop(self) -> Any:
        return next(self)



from clang.cindex import CursorKind, Cursor, Index, TranslationUnit
from clang import cindex

from typing import Any, List

CLANG_INIT = False


def init_clang(_path_to_lib: str):
    global CLANG_INIT
    if not CLANG_INIT:
        cindex.Config.set_library_file(_path_to_lib)
        CLANG_INIT = True


def not_none(value: Any) -> bool:
    return value is not None


def all_as_kinds_generator(kinds: List[CursorKind], cursor: Cursor):
    for element in list(cursor.get_children())[::-1]:
        if element.kind in kinds:
            yield element


def get_translation_unit_index(path_to_file: str, **kwargs) -> TranslationUnit:
    return Index.create().parse(path_to_file, **kwargs)



from clangast.algorithms.elements_parsing import UnitParser
from clangast.algorithms.tools import get_translation_unit_index


def parse_file(path_to_file: str, **kwargs) -> UnitParser:
    parser = UnitParser(get_translation_unit_index(path_to_file, **kwargs))
    parser.parse()
    return parser



from clang.cindex import TranslationUnit, CursorKind
from typing import List, Dict, NoReturn, Generator, Any

from clangast.algorithms.tools.utility import all_as_kinds_generator
from clangast.algorithms.elements_parsing.object_parser import ObjectParser


class UnitParser(object):
    def __init__(self, translation_unit: TranslationUnit):
        self.unit = translation_unit
        self.parsed_objects_index: List[ObjectParser] = []
        self.parsed_objects_reverse_index: Dict[str, int] = {}
        self.target_types: List[CursorKind] = [
            CursorKind.FUNCTION_DECL
        ]

    def append_parsing_target_types(self, *types: List[CursorKind]):
        self.target_types += types

    def set_parsing_target_types(self, *types: List[CursorKind]) -> NoReturn:
        self.target_types = types

    def _all_target_objects_in_translation_unit(self) -> Generator[Any, Any, None]:
        for element in all_as_kinds_generator(self.target_types, self.unit.cursor):
            yield element

    def parse(self) -> NoReturn:
        for element in self._all_target_objects_in_translation_unit():
            if element.spelling in self.parsed_objects_index:
                continue

            parser = ObjectParser(element)
            parser.parse()

            index = len(self.parsed_objects_index)
            self.parsed_objects_index.append(parser)
            self.parsed_objects_reverse_index[parser.name()] = index

    def show(self, target_file, **kwargs) -> NoReturn:
        for parser in self.parsed_objects_index:
            parser.show(target_file, **kwargs)



from setuptools import setup, find_packages

setup(
    name="clangast",
    version="0.2.1.1",
    descriptiom="Find and indexing algorithms for AST based on clang",
    url="https://github.com/DmitIW/clang_ast_wrapper",
    author="Dmitri Ivakhnenko",
    author_email="dmit.ivh@gmail.com",
    license="Apache-2.0",
    packages=find_packages(),
    install_requires=["clang>=11.0"]
)



from clangast.data_structures.generator import StackingPipeline
from clangast.data_structures.node import BaseNode as Node
from clangast.data_structures.tree import BaseTree as Tree

__all__ = [
    "StackingPipeline", "Node", "Tree"
]



from clangast.data_structures.generator import EndlessPipeline
from clangast.algorithms.tools.utility import not_none
from clang.cindex import Token

from typing import Generator, Any

from functools import partial

left_borders = ('(', '[', '{')
right_borders = (')', ']', '}')

borders_index = {
    **dict(zip(left_borders, right_borders)),
    **dict(zip(right_borders, left_borders))
}

filter_none = partial(filter, not_none)


def gathering_tokens_until_border(token_generator: EndlessPipeline,
                                  border_token_spelling: str) -> Generator[Token, Any, None]:
    for token in filter_none(token_generator):
        if token.spelling == border_token_spelling:
            break
        yield token


def gathering_statement_open_spelling(token_generator: EndlessPipeline,
                                      open_border_spelling: str) -> Generator[Token, Any, None]:
    close_border_spelling = borders_index[open_border_spelling]
    return gathering_statement(token_generator, close_border_spelling)


def gathering_statement(token_generator: EndlessPipeline,
                        close_border_spelling: str) -> Generator[Token, Any, None]:
    open_borders = 1
    for token in filter_none(token_generator):
        if open_borders == 1 and not mismatch_spelling(token, close_border_spelling):
            break

        if token.spelling in left_borders:
            open_borders += 1
        elif token.spelling in right_borders:
            open_borders -= 1

        yield token


def mismatch_spelling(token: Token, spelling: str) -> bool:
    return token.spelling == spelling
